﻿using System.Runtime.InteropServices;
using Generator.Common.Exceptions;

namespace Generator.Application.Board
{
    public class Board
    {
        #region Structs
        [StructLayout(LayoutKind.Sequential)]
        private struct SCell
        {
            public int rPos;
            public int cPos;
            public int num;
            public char sType;
        }

        public struct Cell
        {
            public int row;
            public int col;
            public int? num;
            public CellType type;
        }

        #endregion

        #region Enums
        public enum Difficulty
        {
            Easy,
            Medium,
            Hard

        }
        public enum CellType
        {
            White,
            Grey,
            Black
        }

        #endregion

        #region EnumConversions
        public static Difficulty StringToDifficulty(string str)
        {
            return (Difficulty)Enum.Parse(typeof(Difficulty), str);
        }
        public static CellType StringToCellType(string str)
        {
            return (CellType)Enum.Parse(typeof(CellType), str);
        }
        private static CellType CharToCellType(char c)
        {
            return c switch
            {
                'W' => CellType.White,
                'B' => CellType.Black,
                'G' => CellType.Grey,
                _ => CellType.Grey,
            };
        }

        private static char CellTypeToChar(CellType t)
        {
            return t switch
            {
                CellType.White => 'W',
                CellType.Black => 'B',
                CellType.Grey => 'G',
                _ => 'G'
            };
        }

        #endregion

        #region Fields
        private readonly IntPtr _SBoardPtr;

        private readonly double _Distr;

        private readonly HaskellRuntimeSingleton _Singleton;
        #endregion

        #region Properties
        public int Height { get; }
        public int Width { get; }
        public int Seed { get; }
        public Difficulty Diff { get; }
        public Cell[,] UnsolvedBoardArray { get; private set;  }
        public Cell[,] SolvedBoardArray { get; private set; }

        #endregion

        #region Methods
        public void PrintUnsolved()
        {
            for (int i = 0; i < Height; i++)
            {
                for (int j = 0; j < Width; j++)
                {
                    var cell = UnsolvedBoardArray[i, j];
                    Console.Write($"({cell.row}, {cell.col}): {(cell.num == null ? cell.type.ToString() : cell.num)}  ");
                }
                Console.WriteLine();
            }
        }
        public void PrintSolved()
        {
            for (int i = 0; i < Height; i++)
            {
                for (int j = 0; j < Width; j++)
                {
                    var cell = SolvedBoardArray[i, j];
                    Console.Write($"({cell.row}, {cell.col}): {(cell.num == null ? cell.type.ToString() : cell.num)}  ");
                }
                Console.WriteLine();
            }
        }
        private void FillUnsolvedBoard()
        {
            for (int i = 0; i < Height; i++)
            {
                for (int j = 0; j < Width; j++)
                {
                    IntPtr sCellPtr = HaskellRuntimeSingleton.getUnsolvedSCellPtrByIndex(_SBoardPtr, i, j);
                    if (sCellPtr == IntPtr.Zero)
                        throw new IndexOutOfRangeException($"Index ({i}, {j}) does not exist in board");

                    var sCell = (SCell)Marshal.PtrToStructure(sCellPtr, typeof(SCell));
                    Cell cell = new()
                    {
                        row = sCell.rPos,
                        col = sCell.cPos,
                        num = sCell.num == 0 ? null : sCell.num,
                        type = CharToCellType(sCell.sType)
                    };

                    UnsolvedBoardArray[i, j] = cell;
                }
            }
        }
        private void FillSolvedBoard()
        {
            for (int i = 0; i < Height; i++)
            {
                for (int j = 0; j < Width; j++)
                {
                    IntPtr sCellPtr = HaskellRuntimeSingleton.getSolvedSCellPtrByIndex(_SBoardPtr, i, j);
                    if (sCellPtr == IntPtr.Zero)
                        throw new IndexOutOfRangeException($"Index ({i}, {j}) does not exist in board");

                    var sCell = (SCell)Marshal.PtrToStructure(sCellPtr, typeof(SCell));
                    Cell cell = new()
                    {
                        row = sCell.rPos,
                        col = sCell.cPos,
                        num = sCell.num == 0 ? null : sCell.num,
                        type = CharToCellType(sCell.sType)
                    };

                    SolvedBoardArray[i, j] = cell;
                }
            }
        }
        private void FreeMemory()
        {
            if (_SBoardPtr != IntPtr.Zero)
            {
                Console.WriteLine("Freeing board");
                HaskellRuntimeSingleton.freeSBoard(_SBoardPtr);
            }
        }

        #endregion

        public Board(Difficulty difficulty, int height, int width, int timeoutDurationSeconds = 10, int? seed = null, double distr = 0.75)
        {
            _Singleton = HaskellRuntimeSingleton.Instance;

            Height = height;
            Width = width;

            var rand = new Random();
            Seed = seed is null ? rand.Next() : (int)seed;
            Diff = difficulty;
            _Distr = distr;

            UnsolvedBoardArray = new Cell[Height, Width];
            SolvedBoardArray = new Cell[Height, Width];

            var task = Task.Run(() =>
            {
                var result = Diff switch
                {
                    Difficulty.Easy => HaskellRuntimeSingleton.generateEasyBoard(Height, Width, Seed, _Distr),
                    Difficulty.Medium => HaskellRuntimeSingleton.generateMediumBoard(Height, Width, Seed, _Distr),
                    _ => HaskellRuntimeSingleton.generateEasyBoard(Height, Width, Seed, _Distr),
                };
                return result;
            });

            if (!task.Wait(TimeSpan.FromSeconds(timeoutDurationSeconds)))
            {
                FreeMemory();
                throw new BoardTimeoutException(timeoutDurationSeconds, Seed, Height, Width, Diff.ToString());
            }

            _SBoardPtr = task.Result;

            if (_SBoardPtr == IntPtr.Zero)
            {
                FreeMemory();
                throw new UnsolvableBoardException(Seed, Height, Width, Diff.ToString());
            }

            FillUnsolvedBoard();
            FillSolvedBoard();
        }

        ~Board()
        {
            FreeMemory();
        }
    }
}
