module Board.Generator(
    generateGreyBoard,
    generateAndCheckRandomCompletedBoard,
    placeNumbersSequentially,
    Distribution,
    BoardSize
)where

import System.Random ( StdGen, Random (randomR, random) )

import Board.Cell(Cell(..), CellType(..), Coordinate, cellIsGrey)

import Board.Board(Board)

import Board.Modifier(markCells, placeNumber, removeAllMarkings, unMarkCells, blacken, whiten)

import Region.Region(findAreaNeighbours, isCellPotentialPool, isCellPotentialDryLand, findAllGreyCells)

import Region.Island(Island, findLargestIsland, findAllIslands)

import Region.Water(findAllWaterRegions)

--relative weight of Water cells. must be in the interval [0,1]
type Distribution = Double

--(height, width)
type BoardSize = (Int, Int)

_isValidBoardSize :: BoardSize -> Bool
_isValidBoardSize (x, y)
    | x <= 0 = False
    | y <= 0 = False
    |otherwise = True

_isValidDistribution :: Distribution -> Bool
_isValidDistribution p
    |p < 0 || p > 1 = False
    |otherwise = True


generateGreyBoard :: BoardSize -> Board
generateGreyBoard = _generateHomogenousBoard Grey

_generateHomogenousBoard :: CellType -> BoardSize -> Board
_generateHomogenousBoard _ (0, _) = []
_generateHomogenousBoard _ (_, 0) = []
_generateHomogenousBoard t (height, width) = [[Cell {position = (i, j), number = Nothing, cellType = t}
    | j <- [0..(width - 1)]]
    | i <- [0..(height - 1)]]

_generateRandomCompletedBoard :: BoardSize -> Distribution -> StdGen -> Board
_generateRandomCompletedBoard size distribution gen
    |not $ _isValidBoardSize size = []
    |not $ _isValidDistribution distribution = []
    |otherwise = let
            baseBoard = generateGreyBoard size
            (start, genPassed1) = _getRandomStartingPoint size gen
            startedBoard = blacken baseBoard [start]
            (filledBoard, _) = _fillBoard startedBoard distribution genPassed1
        in
            filledBoard

_getRandomStartingPoint :: BoardSize -> StdGen -> (Coordinate, StdGen)
_getRandomStartingPoint size gen
    |not $ _isValidBoardSize size = ((0, 0), gen)
    |otherwise = let
            (x, genPassed1) = randomR (0, fst size - 1) gen
            (y, genPassed2) = randomR (0, snd size - 1) genPassed1
        in
            ((x, y), genPassed2)

_fillBoard :: Board -> Distribution -> StdGen -> (Board, StdGen)
_fillBoard [] _ gen = ([], gen)
_fillBoard board distr gen
    | not $ null waterRegions = (board, gen)
    | null expandableNeighborhood = (board, gen)
    | otherwise =
        let
            (randomIndex, genPassed1) = if length expandableNeighborhood == 1 then
                    (0, gen)
                else
                    randomR (0, length expandableNeighborhood - 1) gen
            pos = position $ expandableNeighborhood !! randomIndex
            (randomType, genPassed2) = _getRandomCellType gen distr
            cellType
                | isCellPotentialPool board pos = White
                | length expandableNeighborhood == 1 = Black
                | isCellPotentialDryLand board pos = Black
                | otherwise = randomType
            newBoard = markCells cellType board [pos]
        in
            _fillBoard newBoard distr genPassed2

    where
        (waterRegion:waterRegions) = findAllWaterRegions board
        expandableNeighborhood = [c | c <- findAreaNeighbours board (map position waterRegion),
            cellIsGrey c]

_getRandomCellType :: StdGen -> Distribution -> (CellType, StdGen)
_getRandomCellType gen distr = let
        (val, newGen) = random gen :: (Double, StdGen)
    in
        if val <= distr then
            (Black, newGen)
        else
            (White, newGen)

generateAndCheckRandomCompletedBoard :: BoardSize -> Distribution -> StdGen -> Board
generateAndCheckRandomCompletedBoard size distr gen =
    let
        startingBoard = _generateRandomCompletedBoard size distr gen
        noGreyCellsBoard = whiten startingBoard (map position $ findAllGreyCells startingBoard)
        checkedBoard = _checkBoard noGreyCellsBoard size distr gen
    in
        checkedBoard

_checkBoard :: Board -> BoardSize -> Distribution -> StdGen -> Board
_checkBoard [] _ _ _ = []
_checkBoard board size distr gen =
    let
        largestIsland = findLargestIsland board
    in
        if length largestIsland > uncurry min size then
            let
                workingBoard = unMarkCells board (map position largestIsland)
                (filledBoard, newGen) = _fillBoard workingBoard distr gen
                newBoard = whiten filledBoard (map position $ findAllGreyCells filledBoard)
            in
                if newBoard == board then
                    board
                else
                    _checkBoard newBoard size distr newGen
        else
            board

placeNumbersSequentially :: Board -> Int -> Board
placeNumbersSequentially [] _ = []
placeNumbersSequentially board sequenceNum =
    let
        islands = findAllIslands board
    in
        if _isSequenceNumValid islands sequenceNum then
            _placeSequentialNumbers board sequenceNum islands
        else
            []

_placeSequentialNumbers :: Board -> Int -> [Island] -> Board
_placeSequentialNumbers [] _ _ = []
_placeSequentialNumbers board _ [] = board
_placeSequentialNumbers board sequenceNum (head:tail) =
    let
        len = length head
        index = sequenceNum `mod` len
        passedSequenceNum = sequenceNum `div` len
        headCoords = position (head !! index)
        newBoard = placeNumber board len headCoords
    in
        _placeSequentialNumbers newBoard passedSequenceNum tail


_isSequenceNumValid :: [Island] -> Int -> Bool
_isSequenceNumValid [] _ = False
_isSequenceNumValid islands num
    | num < 0 = False
    | otherwise = num <= maxSequenceNum
    where
        maxSequenceNum = product (map length islands) - 1
