{-# LANGUAGE ForeignFunctionInterface #-}

module HaskellInterface where

import ExportTypes

import Board.Board(Board(..))
import Board.Cell(Cell(..), CellType(..))

import Nurikabe.Generator(generateEasySeededBoard, generateMediumSeededBoard, generateHardSeededBoard)

import Foreign.Ptr
import Foreign.C.Types
import Foreign.C.String
import Foreign.Marshal
import Foreign.Storable

import Data.Maybe(fromMaybe)

type SCellBoard = [SCellRow]
type SCellRow = [SCell]
type SRow = Ptr SCell

_sBoardToSCellBoards :: SBoard -> IO (SCellBoard, SCellBoard)
_sBoardToSCellBoards sBrd =
    let
        h = fromIntegral $ sHeight sBrd
        w = fromIntegral $ sWidth sBrd
        unsolvedBoardPtr = sUnsolvedBoard sBrd
        solvedBoardPtr = sSolvedBoard sBrd
    in do
        unsolvedPtrArr <- peekArray h unsolvedBoardPtr
        solvedPtrArr <- peekArray h solvedBoardPtr
        unsolvedSCellBoard <- sequence [peekArray w ptr | ptr <- unsolvedPtrArr]
        solvedSCellBoard <- sequence [peekArray w ptr | ptr <- solvedPtrArr]
        return (unsolvedSCellBoard, solvedSCellBoard)



_pokeBoard :: SBoard -> (SCellBoard, SCellBoard) -> IO ()
_pokeBoard sBrd (unSolvedBoard, []) = _pokeUnsolvedBoard sBrd unSolvedBoard
_pokeBoard sBrd ([], solvedBoard) = _pokeSolvedBoard sBrd solvedBoard
_pokeBoard sBrd (unSolvedBoard, solvedBoard) = do
    _pokeUnsolvedBoard sBrd unSolvedBoard
    _pokeSolvedBoard sBrd solvedBoard

_pokeSolvedBoard :: SBoard -> SCellBoard -> IO()
_pokeSolvedBoard sBrd board = do
    ptrArr <- peekArray (fromIntegral $ sHeight sBrd) (sSolvedBoard sBrd)
    let tups = zip ptrArr board
    sequence_ [uncurry pokeArray tup | tup <- tups]

_pokeUnsolvedBoard :: SBoard -> SCellBoard -> IO()
_pokeUnsolvedBoard sBrd board = do
    ptrArr <- peekArray (fromIntegral $ sHeight sBrd) (sUnsolvedBoard sBrd)
    let tups = zip ptrArr board
    sequence_ [uncurry pokeArray tup | tup <- tups]


_sCellToCell :: SCell -> Cell
_sCellToCell sCell = 
    let
        n = fromIntegral $ num sCell
        t = castCCharToChar $ sType sCell
    in
        Cell{position=(fromIntegral $ rPos sCell, fromIntegral $ cPos sCell),
            number = if n == 0 then Nothing else Just n,
            cellType = case t of
                'W' -> White
                'B' -> Black
                _ -> Grey}

_cellToSCell :: Cell -> SCell
_cellToSCell cell =
    let
        rPos = fromIntegral $ fst $ position cell
        cPos = fromIntegral $ snd $ position cell
        num = fromIntegral $ fromMaybe 0 (number cell)
        sType = castCharToCChar $ case cellType cell of
            White -> 'W'
            Grey -> 'G'
            Black -> 'B'
    in SCell{
        rPos = rPos,
        cPos = cPos,
        num = num,
        sType = sType
    }



_boardToSCellBoard :: Board -> SCellBoard
_boardToSCellBoard board = [[_cellToSCell cell | cell <- row ] | row <- board]

_boardsToNewSBoard :: (Board, Board) -> IO (Ptr SBoard)
_boardsToNewSBoard ([], _) = return nullPtr
_boardsToNewSBoard (_, []) = return nullPtr
_boardsToNewSBoard (unsolvedBoard, solvedBoard) = 
    let
        height = fromIntegral $ length unsolvedBoard
        width = fromIntegral $ length $ head unsolvedBoard
    in do
    unsolvedPtrArr <- sequence [newArray [_cellToSCell cell | cell <- row ] | row <- unsolvedBoard]
    sUnsolvedBoard <- newArray unsolvedPtrArr
    solvedPtrArr <- sequence [newArray [_cellToSCell cell | cell <- row ] | row <- solvedBoard]
    sSolvedBoard <- newArray solvedPtrArr
    let newSBoard = SBoard{sHeight=height, sWidth=width, sUnsolvedBoard=sUnsolvedBoard, sSolvedBoard=sSolvedBoard}
    ptr <- malloc
    poke ptr newSBoard
    return ptr

_sCellBoardToBoard :: SCellBoard -> Board
_sCellBoardToBoard sCellBoard = [[_sCellToCell sCell | sCell <- row] | row <- sCellBoard]



foreign export ccall
    generateEasyBoard :: CInt -> CInt -> CInt -> CDouble -> IO (Ptr SBoard)

generateEasyBoard :: CInt -> CInt -> CInt -> CDouble -> IO (Ptr SBoard)
generateEasyBoard height width seed distr =
    let
        h = fromIntegral height
        w = fromIntegral width
        s = fromIntegral seed
        d = fromRational $ toRational distr
        newBoards = generateEasySeededBoard (h, w) d s
    in 
        _boardsToNewSBoard newBoards

foreign export ccall
    generateMediumBoard :: CInt -> CInt -> CInt -> CDouble -> IO (Ptr SBoard)

generateMediumBoard :: CInt -> CInt -> CInt -> CDouble -> IO (Ptr SBoard)
generateMediumBoard height width seed distr =
    let
        h = fromIntegral height
        w = fromIntegral width
        s = fromIntegral seed
        d = fromRational $ toRational distr
        newBoards = generateMediumSeededBoard (h, w) d s
    in
        _boardsToNewSBoard newBoards

foreign export ccall
    generateHardBoard :: CInt -> CInt -> CInt -> CDouble -> IO (Ptr SBoard)

generateHardBoard :: CInt -> CInt -> CInt -> CDouble -> IO (Ptr SBoard)
generateHardBoard height width seed distr =
    let
        h = fromIntegral height
        w = fromIntegral width
        s = fromIntegral seed
        d = fromRational $ toRational distr
        newBoards = generateHardSeededBoard (h, w) d s
    in
        _boardsToNewSBoard newBoards


foreign export ccall
    freeSBoard :: Ptr SBoard -> IO()

freeSBoard :: Ptr SBoard -> IO()
freeSBoard sBrdPtr = do
    sBrd <- peek sBrdPtr
    let unsolvedBoardPtr = sUnsolvedBoard sBrd
        solvedBoardPtr = sSolvedBoard sBrd
        height = fromIntegral $ sHeight sBrd
    unsolvedBoardArr <- peekArray height unsolvedBoardPtr
    solvedBoardArr <- peekArray height solvedBoardPtr
    _freeSRows unsolvedBoardArr
    _freeSRows solvedBoardArr

_freeSRows :: [SRow] -> IO()
_freeSRows [] = return ()
_freeSRows (head:tail) = do
    _freeSRow head
    _freeSRows tail

_freeSRow :: SRow -> IO()
_freeSRow row = do
    free row

foreign export ccall
    getSolvedSCellPtrByIndex :: Ptr SBoard -> CInt -> CInt -> IO (Ptr SCell)

getSolvedSCellPtrByIndex :: Ptr SBoard -> CInt -> CInt -> IO (Ptr SCell)
getSolvedSCellPtrByIndex sBrdPtr cRow cCol = do
    sBrd <- peek sBrdPtr
    let height = fromIntegral $ sHeight sBrd
        width = fromIntegral $ sWidth sBrd
        r = fromIntegral cRow
        c = fromIntegral cCol
    if r >= height || r < 0 || c >= width || c < 0 then
        return nullPtr
    else do
        boardArr <- peekArray height (sSolvedBoard sBrd)
        let rowPtr = boardArr !! r
        sCell <- peekElemOff rowPtr c
        return $ plusPtr rowPtr $ c*sizeOf sCell

foreign export ccall
    getUnsolvedSCellPtrByIndex :: Ptr SBoard -> CInt -> CInt -> IO (Ptr SCell)

getUnsolvedSCellPtrByIndex :: Ptr SBoard -> CInt -> CInt -> IO (Ptr SCell)
getUnsolvedSCellPtrByIndex sBrdPtr cRow cCol = do
    sBrd <- peek sBrdPtr
    let height = fromIntegral $ sHeight sBrd
        width = fromIntegral $ sWidth sBrd
        r = fromIntegral cRow
        c = fromIntegral cCol
    if r >= height || r < 0 || c >= width || c < 0 then
        return nullPtr
    else do
        boardArr <- peekArray height (sUnsolvedBoard sBrd)
        let rowPtr = boardArr !! r
        sCell <- peekElemOff rowPtr c
        return $ plusPtr rowPtr $ c*sizeOf sCell