module Nurikabe.Generator(
    generateEasySeededBoard,
    generateMediumSeededBoard,
    generateHardSeededBoard
)where

import Board.Cell (Cell(..), cellIsBlack)
import Board.Board(Board, coordinateIsValid)
import Board.Generator (BoardSize, Distribution, generateAndCheckRandomCompletedBoard, placeNumbersSequentially)
import Board.Modifier (blacken, removeAllMarkings)

import Region.Region (isCellPotentialPool, findAreaNeighbours)
import Region.Island (findAllIslands)

import Nurikabe.Nurikabe (easySolve, mediumSolve, hardSolve)

import System.Random (mkStdGen)
import Data.List (sortOn)


data Difficulty = Easy | Medium | Hard
    deriving(Show)
    

generateEasySeededBoard :: BoardSize -> Distribution -> Int -> (Board, Board)
generateEasySeededBoard size distr seed =
    let
        gen = mkStdGen seed
        completeBoard = generateAndCheckRandomCompletedBoard size distr gen
        solvedBoard = _makeSolvable Easy completeBoard [] 0
    in
        (removeAllMarkings solvedBoard, solvedBoard)

generateMediumSeededBoard :: BoardSize -> Distribution -> Int -> (Board, Board)
generateMediumSeededBoard size distr seed =
    let
        gen = mkStdGen seed
        completeBoard = generateAndCheckRandomCompletedBoard size distr gen
        solvedBoard = _makeSolvable Medium completeBoard [] 0
    in
        (removeAllMarkings solvedBoard, solvedBoard)

generateHardSeededBoard :: BoardSize -> Distribution -> Int -> (Board, Board)
generateHardSeededBoard size distr seed =
    let
        gen = mkStdGen seed
        completeBoard = generateAndCheckRandomCompletedBoard size distr gen
        solvedBoard = _makeSolvable Hard completeBoard [] 0
    in
        (removeAllMarkings solvedBoard, solvedBoard)

        
_makeSolvable :: Difficulty -> Board -> [Cell] -> Int -> Board
_makeSolvable _ [] _ _ = []
_makeSolvable diff board [] depth = 
    let
        splits = _findSplits board
    in if null splits then
        fst $ _markSplits diff board splits 0 
    else
        _makeSolvable diff board splits depth
_makeSolvable diff board splits depth =
    let
        (solveAttempt, endSearch) = _markSplits diff board splits depth
    in
        if endSearch then
            solveAttempt
        else
            _makeSolvable diff board splits (depth+1)
        
_markSplits :: Difficulty -> Board -> [Cell] -> Int -> (Board, Bool)
_markSplits _ [] _ _ = ([], True)
_markSplits diff board _ 0 = -- at max depth
    let solution = snd $ _placeNumbers diff board 0
    in (solution, not $ null solution)
_markSplits _ board [] _ = ([], True) -- no possible splits
_markSplits diff board (head:tail) depth = -- not at max depth
    let
        newBoard = _split board head
        newSplits = _findSplits newBoard
        (solveAttempt, branchMaxDepthReached) = _markSplits diff newBoard newSplits (depth-1)
        (restSolveAttempt, restMaxDepthReached) = _markSplits diff board tail depth
    in
        if not $ null solveAttempt then
            (solveAttempt, True)
        else if not $ null restSolveAttempt then
            (restSolveAttempt, True)
        else
            ([], branchMaxDepthReached && restMaxDepthReached)

_placeNumbers :: Difficulty -> Board -> Int -> (Board, Board)
_placeNumbers _ [] _ = ([], [])
_placeNumbers diff board sequenceNum =
    let
        numberedBoard = placeNumbersSequentially board sequenceNum
        unsolvedBoard = removeAllMarkings numberedBoard
        solveAttempt = case diff of
            Easy -> fst $ easySolve unsolvedBoard 0
            Medium -> fst $ mediumSolve unsolvedBoard 0
            Hard -> fst $ hardSolve unsolvedBoard 0

    in
        if null solveAttempt && not (null unsolvedBoard) then
            _placeNumbers diff board (sequenceNum + 1)
        else
            (unsolvedBoard, solveAttempt)
        
_findSplits :: Board -> [Cell]
_findSplits [] = []
_findSplits board = 
    let 
        islands = sortOn length $ [island | island <- findAllIslands board, length island > 1]
        islandCells = concat islands
        splits = reverse [cell | cell <- islandCells, _isValidSplit board cell] 
    in
        splits

_split :: Board -> Cell -> Board
_split [] _ = []
_split board cell
    | not $ coordinateIsValid board (position cell) = board
    | otherwise = blacken board [position cell]

_isValidSplit :: Board -> Cell -> Bool
_isValidSplit [] _ = False 
_isValidSplit board cell = 
    let
        isPool = isCellPotentialPool board (position cell)
        cellWaterNeighbours = [waterCell | waterCell <- findAreaNeighbours board [position cell], cellIsBlack waterCell]
        isDiscontinuous = null cellWaterNeighbours
    in
        not (isPool || isDiscontinuous)

