﻿using Microsoft.AspNetCore.Mvc;
using Generator.Application.Board;
using Generator.Dto;
using Generator.Common.Exceptions;

namespace Generator.Controllers
{
    [ApiController]
    [Route("[controller]")]
    public class GenerateSeededController : BaseGenerateController
    {
        [HttpGet(Name = "GenerateSeeded")]
        public ObjectResult GenerateSeeded([FromQuery] string difficulty, [FromQuery] int height, [FromQuery] int width, [FromQuery] int seed, [FromQuery] int timeoutDurationSeconds = 10)
        {
            return GenerateBoard(difficulty, height, width, seed, timeoutDurationSeconds); 
        }

    }

    [ApiController]
    [Route("[controller]")]
    public class GenerateUnseededController : BaseGenerateController
    {
        [HttpGet(Name = "GenerateUnseeded")]
        public ObjectResult GenerateUnseeded([FromQuery] string difficulty, [FromQuery] int height, [FromQuery] int width, [FromQuery] int timeoutDurationSeconds = 10)
        {
            var tokenSource = new CancellationTokenSource();
            var cancellationToken = tokenSource.Token;

            var task = Task.Run(() =>
            {
                while (true)
                {
                    if (cancellationToken.IsCancellationRequested)
                        return null;

                    var result = GenerateBoard(difficulty, height, width, null, timeoutDurationSeconds);

                    if (result.StatusCode == 200)
                        return result;
                }
            }, cancellationToken);

            if (!task.Wait(TimeSpan.FromSeconds(timeoutDurationSeconds)))
            {
                tokenSource.Cancel();
                return StatusCode(500, new { error = $"Unable to generate solvable ({height}, {width}) board of {difficulty} difficulty. Execution timed out after {timeoutDurationSeconds} seconds." });

            }
            return task.Result;
        }
    }

    [ApiController]
    [Route("[controller]")]
    public class GenerateMultipleBoardsController : BaseGenerateController
    {
        [HttpGet(Name = "GenerateMultipleBoards")]
        public ObjectResult GenerateMultiple([FromQuery] string difficulty, [FromQuery] int height, [FromQuery] int width, [FromQuery] int number = 5, [FromQuery] int timeoutDurationSeconds = 10)
        {
            return GenerateMultipleBoards(number, difficulty, height, width, timeoutDurationSeconds);
        }
    }


    public class BaseGenerateController : ControllerBase
    {
        protected ObjectResult GenerateBoard(string difficulty, int height, int width, int? seed = null, int timeoutDurationSeconds = 10)
        {

            try
            {
                var board = new Board(Board.StringToDifficulty(difficulty), height, width, timeoutDurationSeconds, seed);

                return StatusCode(200, new BoardDto(board.Seed, board.Height, board.Width, board.UnsolvedBoardArray, board.SolvedBoardArray));
            }
            catch (UnsolvableBoardException ex)
            {
                Console.WriteLine(ex.Message);
                return StatusCode(500, new { error = ex.Message });
            }
            catch (BoardTimeoutException ex)
            {
                Console.WriteLine(ex.Message);
                return StatusCode(500, new { error = ex.Message });
            }
        }

        protected ObjectResult GenerateMultipleBoards(int number, string difficulty, int height, int width, int timeoutDurationSeconds = 10)
        {
            var tokenSource = new CancellationTokenSource();
            var cancellationToken = tokenSource.Token;

            var results = new List<BoardDto>();

            var tasks = new List<Task<Board?>>();
            for(int i = 0; i < number; i++)
            {
                tasks.Add(Task.Run(() =>
                {
                    while (true)
                    {
                        if (cancellationToken.IsCancellationRequested)
                            return null;
                        try
                        {
                            return new Board(Board.StringToDifficulty(difficulty), height, width, timeoutDurationSeconds);
                        }
                        catch (UnsolvableBoardException ex)
                        {
                            Console.WriteLine(ex.Message);
                        }
                        catch (BoardTimeoutException ex)
                        {
                            Console.WriteLine(ex.Message);
                        }
                    }
                }));
            }

            if(!Task.WaitAll(tasks.ToArray(), TimeSpan.FromSeconds(timeoutDurationSeconds)))
                tokenSource.Cancel();

            foreach (var task in tasks)
            {
                if (task.IsCompletedSuccessfully && task.Result != null)
                {
                    results.Add(new BoardDto(task.Result.Seed, task.Result.Height, task.Result.Width, task.Result.UnsolvedBoardArray, task.Result.SolvedBoardArray));
                }
            }

            if(!results.Any())
                return StatusCode(500, new { });

            return StatusCode(200, results);
        }
    }
}
