﻿using MasterGame.Common.Dtos;
using System;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Text.Json;

namespace MasterGame.Models.Boards
{
    public class Boards
    {
        #region Structs
        public struct Cell
        {
            public int RowPosition;
            public int ColPosition;
            public CellTypeEnum Type;
            public int? Number;
        }

        #endregion

        #region Enums
        public enum DifficultyEnum
        {
            Easy,
            Medium,
            Hard
        }
        public enum CellTypeEnum
        {
            Numbered,
            White,
            Black,
            Grey
        }

        #endregion

        #region Fields
        private readonly GeneratorClient _generatorClient;

        private readonly static string _fileNamespace = "MasterGame.Assets.Boards.";

        private readonly static string _fileExtension = ".json";

        #endregion

        #region Properties
        public int Width { get; }
        public int Height { get; }
        public int Seed { get; }
        public DifficultyEnum Difficulty { get; }
        public Cell[,] UnsolvedBoardArray { get; }
        public Cell[,] SolvedBoardArray { get; }

        #endregion

        #region Private Methods
        private void FillArrays(NurikabeBoardDto boards)
        {
            for (int i = 0; i < Height; i++)
            {
                for (int j = 0; j < Width; j++)
                {
                    try
                    {
                        UnsolvedBoardArray[i, j] = FromString(i, j, boards.UnsolvedBoardArray[i][j]);
                    }
                    catch (IndexOutOfRangeException ex)
                    {
                        Console.WriteLine(ex.Message);
                        UnsolvedBoardArray[i, j] = new Cell() { RowPosition = i, ColPosition = j, Number = null, Type = CellTypeEnum.Grey};
                    }

                    try
                    {
                        SolvedBoardArray[i, j] = FromString(i, j, boards.SolvedBoardArray[i][j]);
                    }
                    catch (IndexOutOfRangeException ex)
                    {
                        Console.WriteLine(ex.Message);
                        SolvedBoardArray[i, j] = new Cell() { RowPosition = i, ColPosition = j, Number = null, Type = CellTypeEnum.Grey };
                    }

                }
            }
        }

        private Cell FromString(int row, int col, string str)
        {
            if (int.TryParse(str, out int number))
            {
                return new Cell()
                {
                    RowPosition = row,
                    ColPosition = col,
                    Number = number,
                    Type = CellTypeEnum.Numbered
                };
            }
            else
            {
                return new Cell()
                {
                    RowPosition = row,
                    ColPosition = col,
                    Number = null,
                    Type = FromString(str)
                };
            }
        }

        private CellTypeEnum FromString(string str)
        {
            return (CellTypeEnum)Enum.Parse(typeof(CellTypeEnum), str);
        }

        private static List<NurikabeBoardDto> ReadDtos(string difficulty)
        {
            var filename = difficulty.ToLower();
            Assembly asm = Assembly.GetExecutingAssembly();
            Stream stream = asm.GetManifestResourceStream(_fileNamespace + filename + _fileExtension);

            try
            {
                using(var reader = new StreamReader(stream))
                {
                    return JsonSerializer.Deserialize<List<NurikabeBoardDto>>(reader.ReadToEnd());
                }
            }
            catch(Exception ex)
            {
                Console.WriteLine(ex.Message);
                return null;
            }
        }

        #endregion

        #region Public Methods
        public int? UnsolvedGetNumber(int row, int col)
        {
            var cell = UnsolvedBoardArray[row, col];
            return cell.Number;
        }
        public CellTypeEnum UnsolvedGetString(int row, int col)
        {
            return UnsolvedBoardArray[row, col].Type;
        }
        public int? SolvedGetNumber(int row, int col)
        {
            var cell = SolvedBoardArray[row, col];
            return cell.Number;
        }
        public CellTypeEnum SolvedGetString(int row, int col)
        {
            return SolvedBoardArray[row, col].Type;
        }

        public void UpdateUnsolvedBoard(CellTypeEnum newType, int row, int col)
        {
            UnsolvedBoardArray[row, col].Type = newType;
        }

        public bool IsSolved()
        {
            bool solved = true;
            for(int i = 0; i < Height; i++)
            {
                for(int j = 0; j < Width; j++)
                {
                    var unsolvedCell = UnsolvedBoardArray[i, j];
                    var solvedCell = SolvedBoardArray[i, j];

                    if (!unsolvedCell.Equals(solvedCell))
                    {
                        if(!(unsolvedCell.Type == CellTypeEnum.Grey && solvedCell.Type == CellTypeEnum.White))
                            solved = false;
                    }
                }
            }
            return solved;
        }

        #endregion

        #region Static Methods
        public static DifficultyEnum StringToDifficulty(string diff)
        {
            return (DifficultyEnum)Enum.Parse(typeof (DifficultyEnum), diff);
        }

        public static int GetNumBoards(string diff)
        {
            return ReadDtos(diff).Count;
        }
        #endregion

        public Boards(DifficultyEnum difficulty, int height, int width, int timeoutDurationSeconds = 10, int? seed = null)
        {
            Difficulty = difficulty;
            Height = height;
            Width = width;

            UnsolvedBoardArray = new Cell[Height, Width];
            SolvedBoardArray = new Cell[Height, Width];

            _generatorClient = GeneratorClient.Instance;

            var task = _generatorClient.generateBoards(Height, Width, Difficulty.ToString(), timeoutDurationSeconds, seed);
            task.Wait();

            var boards = task.Result;

            Seed = boards.Seed;

            FillArrays(boards);

        }

        public Boards(string difficulty, int index)
        {
            Difficulty = StringToDifficulty(difficulty);
            try
            {
                var boardDto = ReadDtos(difficulty)[index];


                Height = boardDto.Height;
                Width = boardDto.Width;

                UnsolvedBoardArray = new Cell[Height, Width];
                SolvedBoardArray = new Cell[Height, Width];

                Seed = boardDto.Seed;

                FillArrays(boardDto);
            }
            catch(IndexOutOfRangeException ex)
            {
                Console.WriteLine(ex.Message);
                return;
            }
        }
    }
}
