﻿using MasterGame.Models.Cells;
using System;
using Xamarin.Essentials;
using Xamarin.Forms;

namespace MasterGame.Models.Boards
{
    public class NurikabeBoard : Grid
    {
        #region Properties
        public int BoardHeight { get; }
        public int BoardWidth { get; }
        public double CellWidth { get; }
        public double CellHeight { get; }

        #endregion

        #region Fields
        private Boards boards;

        #endregion

        #region Methods
        private void _solveStep(int row, int col, Boards.CellTypeEnum newType)
        {
            boards.UpdateUnsolvedBoard(newType, row, col);
        }

        public async void CheckIfSolved()
        {
            if (boards.IsSolved())
            {
                await Application.Current.MainPage.DisplayAlert("CONGRATULATIONS!", "You solved the puzzle!", "OK");
                (Application.Current as App).ClearBoard();
                await Shell.Current.Navigation.PopToRootAsync();
            }
        }

        private void FillBoard()
        {
            for (int i = 0; i < BoardHeight; i++)
            {
                for (int j = 0; j < BoardWidth; j++)
                {
                    var number = boards.UnsolvedGetNumber(i, j);
                    Children.Add(new NurikabeCell((i, j), (CellWidth, CellHeight), this, number), j, i);
                }
            }
        }

        #endregion

        public NurikabeBoard(string difficulty, int height, int width, double margin)
        {
            BoardHeight = height;
            BoardWidth = width;
            Margin = margin;
            boards = new Boards(Boards.StringToDifficulty(difficulty), height, width);

            CellWidth = (DeviceDisplay.MainDisplayInfo.Width / DeviceDisplay.MainDisplayInfo.Density - (Margin.HorizontalThickness / 2)) / BoardWidth;
            CellHeight = CellWidth;


            BackgroundColor = Color.Black;
            Padding = 2;
            ColumnSpacing = 2;
            RowSpacing = 2;

            HorizontalOptions = LayoutOptions.Center;
            VerticalOptions = LayoutOptions.CenterAndExpand;

            for (int i = 0; i < BoardHeight; i++)
            {
                RowDefinitions.Add(new RowDefinition() { Height = GridLength.Star });
            }

            for (int i = 0; i < BoardWidth; i++)
            {
                ColumnDefinitions.Add(new ColumnDefinition() { Width = GridLength.Star });
            }

            FillBoard();
        }

        public NurikabeBoard(string difficulty, double margin, int index = 0)
        {
            try
            {
                boards = new Boards(difficulty, index);
            }
            catch (IndexOutOfRangeException ex)
            {
                Console.WriteLine(ex.Message);
            }

            BoardHeight = boards.Height;
            BoardWidth = boards.Width;

            Margin = margin;

            CellWidth = (DeviceDisplay.MainDisplayInfo.Width / DeviceDisplay.MainDisplayInfo.Density - (Margin.HorizontalThickness / 2)) / BoardWidth;
            CellHeight = CellWidth;

            BackgroundColor = Color.Black;
            Padding = 2;
            ColumnSpacing = 2;
            RowSpacing = 2;

            HorizontalOptions = LayoutOptions.Center;
            VerticalOptions = LayoutOptions.CenterAndExpand;

            for (int i = 0; i < BoardHeight; i++)
            {
                RowDefinitions.Add(new RowDefinition() { Height = GridLength.Star });
            }

            for (int i = 0; i < BoardWidth; i++)
            {
                ColumnDefinitions.Add(new ColumnDefinition() { Width = GridLength.Star });
            }

            for (int i = 0; i < BoardHeight; i++)
            {
                for (int j = 0; j < BoardWidth; j++)
                {
                    var number = boards.UnsolvedGetNumber(i, j);
                    Children.Add(new NurikabeCell((i, j), (CellWidth, CellHeight), this, number), j, i);

                }
            }
        }

        private class NurikabeCell : BoxCell
        {
            public Boards.CellTypeEnum CellType;
            public int? Number { get; set; }

            private readonly Cells.ImageCell _dot;

            private bool _isDotAdded = false;

            private readonly LabelCell _label;

            private readonly NurikabeBoard _parent;

            public NurikabeCell((int row, int col) position, (double x, double y) size, NurikabeBoard parent = null, int? number = null, int? fontsize = null) : base(position, size)
            {
                _parent = parent;
                Number = number;

                if (number != null)
                {
                    CellType = Boards.CellTypeEnum.Numbered;
                    IsVisible = false;

                    _label = new LabelCell(position, size, number.ToString(), Color.Black, fontsize)
                    {
                        InputTransparent = true,
                    };
                    _parent.Children.Add(_label, Position.Col, Position.Row);
                }
                else
                {
                    CellType = Boards.CellTypeEnum.Grey;
                    _dot = new Cells.ImageCell(position, size, ImageSource.FromFile("dot.png"))
                    {
                        InputTransparent=true,
                        IsVisible = false,
                    };
                }
            }

            protected override void OnTapped()
            {
                Boards.CellTypeEnum newType = CellType;
                bool update = false;

                switch (CellType)
                {
                    case Boards.CellTypeEnum.White:
                        newType = Boards.CellTypeEnum.Grey;
                        EnableDot(false);
                        update = true;
                        break;

                    case Boards.CellTypeEnum.Grey:
                        newType = Boards.CellTypeEnum.Black;
                        BackgroundColor = Color.Black;
                        update = true;
                        break;

                    case Boards.CellTypeEnum.Black:
                        newType = Boards.CellTypeEnum.Grey;
                        BackgroundColor = Color.White;
                        update = true;
                        break;
                    
                    default:
                        break;
                }

                if (update)
                {
                    _parent._solveStep(Position.Row, Position.Col, newType);
                    CellType = newType;
                }

                _parent.CheckIfSolved();
            }

            protected override void OnLongPressed()
            {
                Boards.CellTypeEnum newType = CellType;
                bool update = false;

                switch (CellType)
                {
                    case Boards.CellTypeEnum.White:
                        newType = Boards.CellTypeEnum.Grey;
                        EnableDot(false);
                        update = true;
                        break;

                    case Boards.CellTypeEnum.Grey:
                        newType = Boards.CellTypeEnum.White;
                        EnableDot(true);
                        update = true;
                        break;

                    case Boards.CellTypeEnum.Black:
                        newType = Boards.CellTypeEnum.White;
                        EnableDot(true);
                        BackgroundColor = Color.White;
                        update = true;
                        break;

                    default:
                        break;
                }

                if (update)
                {
                    _parent._solveStep(Position.Row, Position.Col, newType);
                    CellType = newType;
                }

                _parent.CheckIfSolved();
            }

            private void EnableDot(bool state)
            {
                if (!_isDotAdded)
                {
                    _parent.Children.Add(_dot, Position.Col, Position.Row);
                    _isDotAdded = true;
                }

                _dot.IsVisible = state;
            }

        }
    }
}
